% shannonEntropy(inputData, frameTimes, inputStride, endFrame, params);
%
%%%%%%%%%%%%%%%%%%%%%% Shannon-entropy calculator %%%%%%%%%%%%%%%%%%%%%%%%%
% <><><><><><><><><><><><><><><><> INPUTS <><><><><><><><><><><><><><><><> 
% ______Input examples______
% input_stride = 20; % Duration of time in seconds (s) for which the frames 
%   are agregated into 1 entropy image.
% endNum = 0; % 0 == all frames.
% columnBlock = 1; % '-1' is a special case which makes the columnBlock 
%   equal to the total number of columns of the image only relevant 
%   for method 1.
% frame_limit_per_iteration = 200; % 0 results in no sectioning so that all 
%   frames (still limited by endNum) are looked at per iteration.
%   If -1 entered into "frame_limit_per_iteration" or 2 into process_method, 
%   will revert into the average number of whole frames per stride time.
% process_method = 2; % 1 or 2 currently excepted. use_raw_image = true. 
%   Automaticly will revert to false if using 'process_method == 1'.
% =========================================================================
% _____REQUIRED_____
%
% inputData:   
%   -String of the folder directory for an image folder OR a dataStructure 
%   containing images as matixis, produced using imgBinStruct.
%
% frameTimes:  
%   -String of the file directory for the relevant excel spreadsheet 
%   formatt of the relevant frame time table. 
%
% inputStride: 
%   -Length of time in secounds, across which the frames will form 1 entropy 
%   image.
%
% endFrame:    
%   -The frame up to which frame will be processed into entropy images. 
%   If 0, all frames will be used.
% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% _____OPTIONAL(params)_____
%
% params.method:               
%   -Either 1 or 2 to select between methods. Default is 2.
%
% params.columnBlock:          
%   -The amount of columns used from each frame per itteration.
%   '-1','0' = special cases which makes the columnBlock equal to the total 
%   number of columns of the image.
%
% params.iterationFrameLimit:  
%   -Limits the number of frames procest per iteration. Rounds to the 
%   nearest whole multiple of average/maximum frames in a stride. 
% =========================================================================
% <><><><><><><><><><><><><><><><> OUTPUTS <><><><><><><><><><><><><><><><>
%
% ____Primary____
% EntropyImages: 
%   List of entopy images as 2D matixes in a cell array.
%
% ____Secondary(optional)____
% Info: 
%   Information regarding the individual entropy images e.g. frames
%   used, time duration, ect...
%
% =========================================================================
% =========================================================================

function [EntropyImages, Info] = shannonEntropy(inputData, frameTimes, inputStride, endFrame, params)

program_start = tic;
%%%%%%%%%%%%%%%%%% Inputs pre-processing %%%%%%%%%%%%%%%%

useRawImg = false;
processMethod = 2;
colBlock = 20;
frameLimitPerIteration = 200;
endNum = 0;
strideInput = 20;
FrameRateTable = frameTimes;

if ischar(inputData)
    useRawImg = true;
end

if nargin >= 5
    if isfield(params, 'method')
        processMethod = params.method;
    end

    if isfield(params, 'columnBlock')
        colBlock = params.columnBlock;
    end
    
    if isfield(params, 'iterationFrameLimit')
        frameLimitPerIteration = params.iterationFrameLimit;
    end
end

if nargin >= 4
    endNum = endFrame;
end
if nargin >= 3
    strideInput = inputStride;
end

if processMethod == 2
    frameLimitPerIteration = -1;
elseif processMethod == 1
    useRawImg = false;
end


file_length = 0;
%%%%%%%%<>%%%%%%%%<>%%%%%%%%<>%%%%%%%% method 2
if useRawImg
    myFolder = inputData;
    filePattern = fullfile(myFolder, '*.tif');
    theFilesList = dir(filePattern);
    numb_of_files = length(theFilesList(:));
    img01 = imread(fullfile(myFolder,theFilesList(1).name));
    if size(img01,3)==3
        gry01 = rgb2gray(img01);
    else
        gry01 = img01;
    end
    [Irows,Icols] = size(gry01);
    Isize = Irows*Icols;
    clear img01 gry01;
%%%%%%%%<>%%%%%%%%<>%%%%%%%%<>%%%%%%%%
else
%%%%%%%%<>%%%%%%%%<>%%%%%%%%<>%%%%%%%% method 1
    Irows = inputData.binned.rows;
    Icols = inputData.binned.cols;
    Isize = inputData.binned.size;
    numb_of_files = inputData.count;
end
%%%%%%%%<>%%%%%%%%<>%%%%%%%%<>%%%%%%%%

frameTimeTable = readtable(FrameRateTable);
frametime = zeros(2,1);
if (endNum == 0)
%     endNum = length(theFilesList);
    endNum = numb_of_files;
    frametime(:) = table2array(frameTimeTable(end,2:3));
    frames = frametime(2);
else
    frames = endNum;
    frametime = table2array(frameTimeTable(frames,2:3));
end

msLength = frametime(1);
sLength = msLength/1000;

frameTimeArray = zeros(frames,4);
frameTimeArray(:,1) = table2array(frameTimeTable(1:frames,3)); % frame number
frameTimeArray(:,2) = table2array(frameTimeTable(1:frames,2)); % time (ms)
frameTimeArray(:,3) = frameTimeArray(:,2)/1000; % time (s) 


%%%%%%%%%%%%%%%%% Preprocessing %%%%%%%%%%%%%%%%%
%%%%%%%%
if strideInput == 0
    stride = 20; % i.e. using a stride of 20 s
else
    stride = strideInput; % i.e. using a stride of 20 s
end
upper = stride;
lower = 0;
sections = round(sLength/stride);
for i = 1:sections
    if i~=sections
        frameTimeArray(frameTimeArray(:,3)>lower & frameTimeArray(:,3)<=upper,4) = i;
    else
        frameTimeArray(frameTimeArray(:,3)>lower,4) = i;
    end
    lower = upper;
    upper = upper + stride;
end
% section_frames = zeros(sections,1,'uint8');
section_frames = zeros(sections,1);
for i = 1:sections
    section_frames(i) = length(frameTimeArray(frameTimeArray(:,4) == i));
end
%%%%%%%%
frameStride_max = max(section_frames(:));
%%%%%%%%

if frameLimitPerIteration == -1
    frameLimitPerIteration = frameStride_max;
elseif frameLimitPerIteration == 0
    frameLimitPerIteration = frames;
end

strides_per_iteration_max = 1;

if frameLimitPerIteration == frames
    strides_per_iteration_max = sections;
end

if frameLimitPerIteration < frameStride_max
    strides_per_iteration_max = 1;
    frameLimitPerIteration = frameStride_max;
elseif frameLimitPerIteration > frameStride_max
    strides_per_iteration_max = floor(frameLimitPerIteration/frameStride_max);
    frameLimitPerIteration = strides_per_iteration_max*frameStride_max; 
end
%%%%%%%%<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% strides_per_iteration_max:-   determines the number of frames (from the number of strides per iteration) 
%                               entered into the tabel for calculating pixel
%                               entropys during each itteration and therfore
%                               for how many images (each made up of frames from the stride duration)
%                               pixel entropys are calculated for each iteration.
%
% frame_limit_per_iteration:-   maximum number of frames entered into a
%                               tabel during an iteration.
%
% frameStride_max:-             the maximum number of frames allowed per stride.
%
% stride:-                      the length of time in secounds for each section the video is
%                               divided into (the number of frames should be roughly equal between each
%                               stride).
%
% sections:-                    the numer of strides being used in total.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>

itteration_sections = ceil(sections/strides_per_iteration_max);




%%%%%%%%<>%%%%%%%%<>%%%%%%%%<>%%%%%%%% method 1
if colBlock == 0
    colBlock = 1;
elseif colBlock == -1
    colBlock = Icols;
end

if colBlock > 96
    while rem(Icols,colBlock) ~= 0
        colBlock = colBlock - 1;
    end
else
    while rem(Icols,colBlock) ~= 0
        colBlock = colBlock + 1;
    end
end
%%%%%%%%<>%%%%%%%%<>%%%%%%%%<>%%%%%%%%

% >>>>>>>>>>>>>>>>> Info array building <<<<<<<<<<<<<<<<<<
Info = struct;
Info.method = processMethod;
Info.stride = stride;
Info.framesTotal = frames;
Info.durationTotal = frameTimeArray(end,3);
Info.framesPerIteration = frameLimitPerIteration;
Info.colBlock = colBlock;
% Info: columns |index|start-time(s)|end-time(s)|duration(s)|frames|culmative frames|
infoA = zeros(sections,6);
infoA(:,1) = 1:sections;
infoA(:,5) = section_frames(:);
framePointer = 0;
for i = 1:sections
    framePointer = framePointer + 1; % frame index = (previus cumulative frames + 1) = this section start frame.
    infoA(i,6) = sum(infoA(1:i,5)); % column 6 = cumulative frames (cf).
    infoA(i,2) = frameTimeArray(framePointer,3); % column 2 = frameTimes(sf) = start time.
    framePointer = infoA(i,6); % frame index = this cumulative frames i.e. end frame index for the section.
    infoA(i,3) = frameTimeArray(framePointer,3); % column 3 = frameTimes(cf) = end time.
    infoA(i,4) = infoA(i,3) - infoA(i,2); % column 4 = time diferance = duration. 
end
Info.array = infoA;
% <<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>
%%%%%%%%%%%%%%%%% Main processing step START %%%%%%%%%%%%%%%%%

% method 1+2: initialisations

EntropyImages = cell(sections,1); % 1 + 2.
for i = 1:sections % 1 + 2.
    EntropyImages{i} = zeros(Irows,Icols, 'single'); % 1 + 2.
end % 1 + 2.
two56 = 256; % 1+2
two55 = 255; % 1+2
frame_high = 0; % 1+2
frame_low = 0; % 1+2
stride_idx = 0; % 1+~2


%%%%%%%%%%%%%%%%%%%
% PROCESS METHODS %
%%%%%%%%%%%%%%%%%%%

if processMethod == 2
    % method 2: initialisations 
    % im_cube2 = zeros(Irows,Icols,frameStride_max,'single'); % 2 % may need to change size while processing!
    temp_mat2 = zeros(Irows,Icols,'single'); % 2
    temp_histo2 = zeros(Isize,two56,'single'); % 2
    N2 = 0; % 2
    f1 = 0; % 2
    this_itteration_frames = 0; % 2
    temp_ent2 = zeros(Irows,Icols,'single'); % 2
    temp_prob2 = zeros(Isize,two56,'single'); % 2
    temp_logProb2 = zeros(Isize,two56,'single'); % 2
    temp_probLogProb2 = zeros(Isize,two56,'single'); % 2
    temp_probLogProbSum2 = zeros(Isize,1,'single'); % 2
    % method 2: itterations/process loop
    processing_start = tic;
    for si = 1:itteration_sections
        frame_low = frame_high + 1;
        frame_high = frame_high + section_frames(si);
        
        f1 = 1; % reset for each iteration.
        im_cube2 = zeros(Irows,Icols,section_frames(si),'single');
        if useRawImg
            for fr = frame_low:frame_high
                im_cube2(:,:,f1)= imread(fullfile(myFolder,theFilesList(fr).name));
                f1 = f1 + 1;
            end
        else
            for fr = frame_low:frame_high
                im_cube2(:,:,f1)= inputData.binned_container{fr};
                f1 = f1 + 1;
            end
        end
        
        
       
        for i = 0:two55
            temp_01_cube2 = im_cube2 == i;
            temp_mat2 = sum(temp_01_cube2,3);
            temp_histo2(:,(i+1)) = temp_mat2(:);
        end
        this_itteration_frames = section_frames(si);
        N2 = single(this_itteration_frames);
        temp_prob2 = temp_histo2/N2;
        temp_logProb2 = log(temp_prob2);
        temp_probLogProb2 = temp_prob2.*temp_logProb2;
        temp_probLogProb2(isnan(temp_probLogProb2))=0;
       
        temp_probLogProbSum2 = sum(temp_probLogProb2,2);
        temp_ent2(:) = temp_probLogProbSum2(:);
   
        EntropyImages{si} = temp_ent2*(-1); % convert into negative
        fprintf('Finished processing %i out of %i total sections. \n', si, itteration_sections);
    end
    processing_time = toc(processing_start);
    program_time = toc(program_start);
    %%%%%%%%%%%%%%%%% Main processing step END %%%%%%%%%%%%%%%%%
    fprintf('processing took %d seconds, program took %d seconds. \n', processing_time,program_time);
elseif processMethod == 1
    % method 1: initialisations
    fprintf('working with bolcks of %i columns per iteration. \n', colBlock);
    chunk_nums = Icols/colBlock; % 1
    tabR = Irows * colBlock; % 1
    tabC = frameLimitPerIteration; % 1
    Tsize = tabR*tabC; % 1
    myTab = zeros(tabR,tabC,'uint8'); % 1
    histo = zeros(two56,tabR,'uint8'); % 1
    Intensity_histo = cell(sections,1); % 1.
    for i = 1:sections % 1.
        Intensity_histo{i} = zeros(tabR,two56, 'uint8'); % 1.
    end % 1.
    % --
    temp_prob = zeros(two56,tabR,'single'); % 1
    temp = zeros(two56,tabR,'single'); % 1
    temp_logProb = zeros(two56,tabR,'single'); % 1
    temp_prodLogProb = zeros(two56,tabR,'single'); % 1
    temp_pixel_entropy_col = zeros(tabR,1,'single'); % 1
    stride_idx_max = length(section_frames(:,1)); % 1
    % --
    f1 = 0; % 1+2
    end_iteration_stride_count = 0; % 1
    this_iteration_start_stride = 0; % 1
    this_iteration_end_stride = 0; % 1
    % method 1: itterations/process loop
    processing_start = tic;
    for si = 1 : itteration_sections

        frame_low = frame_high +1;
        this_iteration_start_stride = stride_idx +1;

        if si < itteration_sections
            % strides_per_iteration_max
            for ix = 1:strides_per_iteration_max
                stride_idx = stride_idx + 1;
                frame_high = frame_high + section_frames(stride_idx,1);
            end
        else
            end_iteration_stride_count = (stride_idx_max - stride_idx);

            for ix = 1:end_iteration_stride_count
                stride_idx = stride_idx + 1;
                frame_high = frame_high + section_frames(stride_idx,1);
            end
        end

        this_iteration_end_stride = stride_idx;

        chunk_low = 0;
        chunk_high = 0;
        for ch = 1 : chunk_nums
            chunk_low = chunk_high + 1;
            chunk_high = ch*colBlock*Irows;

            f1 = 1;
            for fr = frame_low:frame_high

                %%%%%%<>%%%%%%<>%%%%%%<>%%%%%%<>%%%%%%<>%%%%%%
    %             img = imread(fullfile(myFolder,theFilesList(fr).name));
    %             if size(img,3)==3
    %                 img = rgb2gray(img);
    %             end
    %             myTab(:,f1) = img(chunk_low:chunk_high);
                %%%%%%<>%%%%%%<>%%%%%%<>%%%%%%<>%%%%%%<>%%%%%%

                myTab(:,f1) = inputData.binned_container{fr}(chunk_low:chunk_high);
                f1 = f1 +1;
            end

            %%%%%%%%%% proces mytab to get entropy values over 20 s (or an input stride) intervals of the column of pixels
            limH = 0;
            limL = 0;
            Ione = 1;
            for i = this_iteration_start_stride:this_iteration_end_stride % sections == total number of stride
                limL = limH+1;
                limH = limH + section_frames(i);
                for j = 0:255
                    j1 = j+1; % +1 adjustment required as the are 256 values between 0:255
                    % therfore index 1 refers to value 0 and 2 to 1 ... 256 to 255
                    %------------------------------------------------------
                    % Original:
                    %   version gives wierd results revert back to this original version
                    % tempTab = (myTab(:,limL:limH)==j)'; % if the below
                    % Intensity_histo{i}(1:end,j1) = sum(tempTab);
                    % - - - - - - - - - - - - - - - - - - - - - - - - - - -
                    % Modified:
                    %   should do as the above original but maybe more efficient % UNTESTED
                    tempTab = (myTab(:,limL:limH)==j); 
                    Intensity_histo{i}(1:end,j1) = sum(tempTab,2);
                    %------------------------------------------------------
                end
                temp = single(Intensity_histo{i}');
                N = single(section_frames(i));
                temp_prob = temp/N;
                temp_logProb = log(temp_prob);
                temp_probLogProb = temp_prob.*temp_logProb;
                temp_probLogProb(isnan(temp_probLogProb))=0;
                temp_pixel_entropy_col = sum(temp_probLogProb);
                %%%%%%%%%% insert each 20 s entropy column into the correct place in
                %%%%%%%%%% the correct entropy image. for ~480 s video will get 24
                %%%%%%%%%% seperate entropy images
                EntropyImages{i}(chunk_low:chunk_high) = temp_pixel_entropy_col(:)*(-1);
            end
        end
    end
    processing_time = toc(processing_start);
    program_time = toc(program_start);
    %%%%%%%%%%%%%%%%% Main processing step END %%%%%%%%%%%%%%%%%
    fprintf('processing took %d seconds, program took %d seconds. \n', processing_time,program_time);
end
fprintf('Average time spent to process 1 frame = %d seconds. \n', processing_time/frames);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% PROGRAM END %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%